
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceContext;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DragSourceMotionListener;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.InvalidDnDOperationException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.SwingUtilities;

import net.miginfocom.swing.MigLayout;

public class DockingHandle extends JComponent {
  private static final long serialVersionUID = 1L;

  private Dockable dockable;
  private final JLabel label;
  private final JButton button;

  private final DragSource ds = new DragSource();
  private final DragHandler dh = new DragHandler();

  private final ActivationHandler ah = new ActivationHandler();

  public DockingHandle(String title) {
    this(title, null);
  }

  public DockingHandle(String title, Dockable dockable) {
    this.dockable = dockable;
    ds.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_MOVE, dh);
    ds.addDragSourceListener(dh);
    ds.addDragSourceMotionListener(dh);

    addMouseListener(ah);

    label = new JLabel(title);
    label.setOpaque(false);

    final ImageIcon icon =
      new ImageIcon(DockingHandle.class.getResource("x.png"));

    button = new JButton(icon);
    button.setOpaque(false);
    button.setBorderPainted(false);
    button.setRolloverEnabled(true);

    final Insets in = new Insets(5,5,5,5);
    final Dimension bs = new Dimension(
      in.left + in.right + icon.getIconWidth(),
      in.top + in.bottom + icon.getIconHeight());

    button.setMinimumSize(bs);
    button.setPreferredSize(bs);
    button.setMaximumSize(bs);

    button.addActionListener(new CloseHandler());

    final MigLayout layout = new MigLayout("insets 0, gap rel:push");
    setLayout(layout);
    add(label, "cell 0 0");
    add(button, "cell 1 0");
  }

  private class CloseHandler implements ActionListener {
    public void actionPerformed(ActionEvent e) {
      final Dock src = dockable.getParentDock();
      src.removeDockable(dockable);

      final Dock root = DockUtils.findRoot(src); 
      DockUtils.collapseTabDocks(root);
      DockUtils.collapseSplitDocks(root);

      ((Container) root.getComponent()).invalidate();
// FIXME: why is this necessary?
      ((Container) root.getComponent()).doLayout();
      root.getComponent().repaint();  
    }
  }

  public void setDockable(Dockable dockable) {
    this.dockable = dockable;
  }

  public Dockable getDockable() {
    return dockable;
  }
  
  public String getTitle() {
    return label.getText(); 
  }

  private void makeSelected() {
    final TabDock dock = (TabDock) dockable.getParentDock();
    final Component comp = dockable.getComponent();
    if (dock.getSelectedComponent() != comp) {
      dock.setSelectedComponent(comp);
    }
  }

  private BufferedImage createImage() {
    final TabDock dock = (TabDock) dockable.getParentDock();

    final Rectangle tabBounds =
      dock.getBoundsAt(dock.indexOfTabComponent(this));

    final BufferedImage img = new BufferedImage(
      tabBounds.width, tabBounds.height, BufferedImage.TYPE_INT_ARGB);

    // Note: It appears to be necessary to use print() instead of
    // paint() here, since paint() resutls in a corrupted image.
    final Graphics2D g = img.createGraphics();
    g.translate(-tabBounds.x, -tabBounds.y);
    dock.getComponent().print(g);
    g.dispose();

    return img;
  }

  private class DragHandler implements DragGestureListener,
                                       DragSourceListener,
                                       DragSourceMotionListener {
    private boolean isOutsideApp(Point p) {
      for (Window w : Window.getWindows()) {
        if (w.getBounds().contains(p)) return false;
      } 
      return true;
    }

    //
    // DragGestureListener
    //
    public void dragGestureRecognized(DragGestureEvent e) {
      makeSelected();

      try {
        e.startDrag(DragSource.DefaultMoveDrop,
                    new TransferableDockable(dockable));
      }
      catch (InvalidDnDOperationException ex) {
        ex.printStackTrace();
        return;
      }

      final Point dp = new Point(e.getDragOrigin());
      SwingUtilities.convertPointToScreen(dp, e.getComponent());

      final TabDock dock = (TabDock) dockable.getParentDock();
      final Point off = dock.getBoundsAt(
        dock.indexOfTabComponent(DockingHandle.this)).getLocation();
      SwingUtilities.convertPointToScreen(off, dock);

      off.x = dp.x - off.x;
      off.y = dp.y - off.y;

      DragGlassPane.setImage(createImage(), off);
      DragGlassPane.forDock(dock).setPosition(dp);
      DragGlassPane.showAll();
    }

    //
    // DragSourceListener
    //
    public void dragDropEnd(DragSourceDropEvent e) {
      DragGlassPane.hideAll();

      // create a new window for a drop outside the app
      if (isOutsideApp(e.getLocation())) {
// FIXME: DockWindows should close automatically when empty
        final DockWindow dw = new DockWindow();
        final TabDock td = new TabDock();
        dw.getDock().addDockable(td, null);

        final Dock src = dockable.getParentDock();
        src.removeDockable(dockable);

        final Dock root = DockUtils.findRoot(src); 
        DockUtils.collapseTabDocks(root);
        DockUtils.collapseSplitDocks(root);

        td.addDockable(dockable);
        td.setSelectedComponent(dockable.getComponent());

        dw.pack();
        dw.setVisible(true);
      }
    }

    public void dragEnter(DragSourceDragEvent e) {
      final DragSourceContext ctx = e.getDragSourceContext();
      ctx.setCursor(
        ctx.getTransferable()
           .isDataFlavorSupported(TransferableDockable.FLAVOR) ?
        DragSource.DefaultMoveDrop : DragSource.DefaultMoveNoDrop
      );
    }

    public void dragExit(DragSourceEvent e) {
      e.getDragSourceContext().setCursor(DragSource.DefaultMoveNoDrop);
    }

    public void dragOver(DragSourceDragEvent e) {}

    public void dropActionChanged(DragSourceDragEvent e) {}

    //
    // DragSourceMotionListener
    //
    public void dragMouseMoved(DragSourceDragEvent e) {
// FIXME: cannot show drag cursor outside of app window
      DragGlassPane.updatePositions(e.getLocation());
    }
  }

  private class ActivationHandler implements MouseListener {
    //
    // MouseListener
    //
    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
      // the DragGestureRecognizer seems to eat MouseEvents, so we have to
      // bring tabs to the front manually 
      makeSelected();
    }
  
    public void mouseReleased(MouseEvent e) {
    } 
  }

/*
  private class ExitHandler implements MouseListener {
    public void mouseEntered(MouseEvent e) {
      System.out.print("o");
    }

    public void mouseExited(MouseEvent e) {
      System.out.print("x");
    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
    }
  
    public void mouseReleased(MouseEvent e) {
    } 
  }
*/
}
